(* :Package Version: 2.0.07.27  *)

(* :Copyright: Copyright 2004, Selwyn Hollis *)

(* :Name: RDL`ReactionDiffusionLab` *)
 
(* :Author: Selwyn Hollis *)

(*	New in version 1.5:
	The aternating direction implicit (ADI) method replaces Crank-Nicolson.

	New in version 2.0:
	Support for non-square rectangular domains and periodic boundary conditions.
	Significant speedup by use of SparseArrays and LinearSolveFunctions.
   
 *)

BeginPackage["RDL`ReactionDiffusionLab`"]

Unprotect["RDL`ReactionDiffusionLab`*"];
ClearAll["RDL`ReactionDiffusionLab`*"];

RDDensityPlots::usage =
"RDDensityPlots[{f[u,v], g[u,v]}, {u,v}, {u0,v0}
\n            {d1,d2}, dt, nsteps, options]\nor
\nRDDensityPlots[{f[u,v,w], g[u,v,w]}, {u,v,w}, {u0,v0,w0},
\n            {d1,d2,d3}, dt, nsteps, options].";

RDSurfacePlots::usage =
"RDSurfacePlots[{f[u,v], g[u,v]}, {u,v}, {u0,v0},
\n            {d1,d2}, dt, nsteps, options]\nor
\nRDSurfacePlots[{f[u,v,w], g[u,v,w]}, {u,v,w}, {u0,v0,w0},
\n            {d1,d2,d3}, dt, nsteps, options].";


Options[RDDensityPlots]:= 
	Union[{ReturnLast->False, PlotEvery->1, Domain->{{0,Automatic},{0,1}},
	BoundaryConditions->Neumann, BoundaryFunctions->{0&,0&}, GraphicsSpacing->0.1}, Options[ListDensityPlot] ];
	
SetOptions[RDDensityPlots,Frame->False, Mesh->False, FrameTicks->None,
	ColorFunction->GrayLevel, PlotRange->Automatic, AspectRatio->Automatic];
	
Options[RDSurfacePlots]:=
	Union[{ReturnLast->False, PlotEvery->1, Domain->{{0,Automatic},{0,1}},
	BoundaryConditions->Neumann, BoundaryFunctions->{0&,0&}, GraphicsSpacing->0.1}, Options[ListPlot3D] ];
	
SetOptions[RDSurfacePlots, Mesh->False, ClipFill->None, PlotRange->Automatic, AspectRatio->Automatic];

Begin["`Private`"]

<<Utilities`FilterOptions`;
	 
(* Routines used by RDstep__ functions (defined below) *)

neumannpad= Compile[{{vals,_Real,2}},
	With[{q=(Append[#,Last[#]]&)/@(Prepend[#,First[#]]&/@vals)},
		Append[Prepend[q, First[q]], Last[q]]
		] ];

neumannXhalfstep= Compile[{{vals,_Real,2},{hh,_Real}},
	With[{ker={{0,0,0},{1,-2,1},{0,0,0}}},
		vals + 0.5*hh*ListCorrelate[ker, neumannpad[vals] ]
		] ];
		
neumannYhalfstep= Compile[{{vals,_Real,2},{hh,_Real}},
	With[{ker={{0,1,0},{0,-2,0},{0,1,0}}},
		vals + 0.5*hh*ListCorrelate[ker, neumannpad[vals] ]
		] ];
							
dirichletpad= Compile[{{vals,_Real,2}},
	Module[{q, n=Last[Dimensions[vals]], o}, o=Table[0,{n+2}];
		q=(PadLeft[#,n+2]&)/@(PadRight[#,n+1]&/@vals);
		Append[Prepend[q,o],o]
		] ];

dirichletXhalfstep= Compile[{{vals,_Real,2},{hh,_Real}},
	With[{ker={{0,0,0},{1,-2,1},{0,0,0}}},
		vals + 0.5*hh*dirichletpad[ ListCorrelate[ker,vals] ]
		] ];
		
dirichletYhalfstep= Compile[{{vals,_Real,2},{hh,_Real}},
	With[{ker={{0,1,0},{0,-2,0},{0,1,0}}},
		vals + 0.5*hh*dirichletpad[ ListCorrelate[ker,vals] ]
		] ];

setedges= Compile[{{vals,_Real,2},{bd,_Real,2},{a010,_Integer,2}}, vals*a010 + bd];

periodicize= Compile[{{vals,_Real,2}},
        With[{q=(ReplacePart[#,.33(First[#]+Last[#])+.17(#[[2]]+#[[-2]]), {{1},{-1}} ]&)/@vals},
		ReplacePart[q,.33(First[q]+Last[q])+.17(q[[2]]+q[[-2]]), {{1},{-1}} ]
		] ];
       
periodicXhalfstep= Compile[{{vals,_Real,2},{hh,_Real}},
	vals + 0.5*hh*(ListCorrelate[{ 1, -2, 1 }, #, {2,2} ]& /@ vals) ];
		
periodicYhalfstep= Compile[{{vals,_Real,2},{hh,_Real}},
	vals + 0.5*hh*ListCorrelate[ {{0,1,0},{0,-2,0},{0,1,0}}, vals, {2,2} ] ];

ADIXu[rhs_]:= Partition[
    uLsolvex[ Flatten[rhs] ], n];

ADIYu[rhs_]:= Transpose@Partition[
	uLsolvey[ Flatten[Transpose@rhs] ], m];

ADIXv[rhs_]:= Partition[
	vLsolvex[ Flatten[rhs] ], n];

ADIYv[rhs_]:= Transpose@Partition[
	vLsolvey[ Flatten[Transpose@rhs] ], m];

ADIXw[rhs_]:=Partition[
	wLsolvex[ Flatten[rhs] ], n];

ADIYw[rhs_]:= Transpose@Partition[
	wLsolvey[ Flatten[Transpose@rhs] ], m];


(** Definitions of RDstep__ functions. 
    There's one for each combination of boundary conditions. **)

RDstepNN[{u0_, v0_}, {fc_,gc_}, {d1_,d2_}, {hhx_,hhy_}]:= 
	Module[{u1,v1},
		u1= ADIXu@ (neumannYhalfstep[u0, d1*hhy ] + fc[u0,v0] );
		v1= ADIXv@ (neumannYhalfstep[v0, d2*hhy ] + gc[u0,v0] );
		{   ADIYu@ (neumannXhalfstep[u1, d1*hhx ] + fc[u1,v1] ),
		    ADIYv@ (neumannXhalfstep[v1, d2*hhx ] + gc[u1,v1] )}
		 ];
		 
RDstepPeriodic[{u0_, v0_}, {fc_,gc_}, {d1_,d2_}, {hhx_,hhy_}]:= 
	Module[{u1,v1},
		u1= ADIYu@(periodicXhalfstep[u0, d1*hhx ] + fc[u0,v0] );
		v1= ADIXv@(periodicYhalfstep[v0, d2*hhy ] + gc[u0,v0] );
	 	{  ADIXu@ (periodicYhalfstep[u1, d1*hhy ] + fc[u1,v1] ),
		   ADIYv@ (periodicXhalfstep[v1, d2*hhx ] + gc[u1,v1] ) }
		 ];

RDstepDD[{u0_, v0_}, {fc_,gc_}, {d1_,d2_}, {hhx_,hhy_}]:= 
	Module[{u1,v1},
		u1= ADIXu@ (dirichletYhalfstep[u0, d1*hhy ] + fc[u0,v0] );
		v1= ADIXv@ (dirichletYhalfstep[v0, d2*hhy ] + gc[u0,v0] );
		{setedges[ ADIYu@ (dirichletXhalfstep[u1, d1*hhx ] + fc[u1,v1]), First[bdata], a010],
		 setedges[ ADIYv@ (dirichletXhalfstep[v1, d2*hhx ] + gc[u1,v1]), Last[bdata],  a010]}];
		 
RDstepND[{u0_, v0_}, {fc_,gc_}, {d1_,d2_}, {hhx_,hhy_}]:= 
	Module[{u1,v1},
		u1= ADIXu@ (  neumannYhalfstep[u0, d1*hhy ] + fc[u0,v0] );
		v1= ADIXv@ (dirichletYhalfstep[v0, d2*hhy ] + gc[u0,v0] );
		{          ADIYu@ ( neumannXhalfstep[u1, d1*hhx ] + fc[u1,v1]),
		 setedges[ ADIYv@ (dirichletXhalfstep[v1, d2*hhx] + gc[u1,v1]), Last[bdata], a010] }];

RDstepDN[{u0_, v0_}, {fc_,gc_}, {d1_Real,d2_Real}, {hhx_,hhy_}]:= 
	Module[{u1,v1},
		u1= ADIXu@ (dirichletYhalfstep[u0, d1*hhy] + fc[u0,v0] );
		v1= ADIXv@ (  neumannYhalfstep[v0, d2*hhy] + gc[u0,v0] );
		{setedges[ ADIYu@ (dirichletXhalfstep[u1, d1*hhx ] + fc[u1,v1]), First[bdata], a010],
		           ADIYv@ (  neumannXhalfstep[v1, d2*hhx ] + gc[u1,v1])} ];
		 
RDstepNNN[{u0_, v0_, w0_}, {fc_,gc_,hc_}, {d1_,d2_,d3_}, {hhx_,hhy_}]:= 
	Module[{u1,v1,w1},
		u1= ADIXu@ (neumannYhalfstep[u0, d1*hhy] + fc[u0,v0,w0] );
		v1= ADIXv@ (neumannYhalfstep[v0, d2*hhy] + gc[u0,v0,w0] );
		w1= ADIXw@ (neumannYhalfstep[w0, d3*hhy] + hc[u0,v0,w0] );
		{   ADIYu@ (neumannXhalfstep[u1, d1*hhx] + fc[u1,v1,w1] ),
		    ADIYv@ (neumannXhalfstep[v1, d2*hhx] + gc[u1,v1,w1] ),
		    ADIYw@ (neumannXhalfstep[w1, d3*hhx] + hc[u1,v1,w1] ) }
		 ];
		 
RDstepNND[{u0_, v0_, w0_}, {fc_,gc_,hc_}, {d1_,d2_,d3_}, {hhx_,hhy_}]:= 
	Module[{u1,v1,w1},
		u1= ADIXu@ (  neumannYhalfstep[u0, d1*hhy ] + fc[u0,v0,w0] );
		v1= ADIXv@ (  neumannYhalfstep[v0, d2*hhy ] + gc[u0,v0,w0] );
		w1= ADIXw@ (dirichletYhalfstep[w0, d3*hhy ] + hc[u0,v0,w0] );
		{			ADIYu@ (  neumannXhalfstep[u1, d1*hhx ] + fc[u1,v1,w1] ),
		 			ADIYv@ (  neumannXhalfstep[v1, d2*hhx ] + gc[u1,v1,w1] ),
		  setedges[ ADIYw@ (dirichletXhalfstep[w1, d3*hhx ] + hc[u1,v1,w1] ), Last[bdata], a010]}
		 ];

RDstepNDD[{u0_, v0_, w0_}, {fc_,gc_,hc_}, {d1_,d2_,d3_}, {hhx_,hhy_}]:= 
	Module[{u1,v1,w1},
		u1= ADIXu@ (  neumannYhalfstep[u0, d1*hhy ] + fc[u0,v0,w0] );
		v1= ADIXv@ (dirichletYhalfstep[v0, d2*hhy ] + gc[u0,v0,w0] );
		w1= ADIXw@ (dirichletYhalfstep[w0, d3*hhy ] + hc[u0,v0,w0] );
		{          ADIYu@ (neumannXhalfstep[  u1, d1*hhx ] + fc[u1,v1,w1] ),
		 setedges[ ADIYv@ (dirichletXhalfstep[v1, d2*hhx ] + gc[u1,v1,w1] ), bdata[[2]],  a010],
		 setedges[ ADIYw@ (dirichletXhalfstep[w1, d3*hhx ] + hc[u1,v1,w1] ), Last[bdata], a010]}
		 ];
		 
RDstepDND[{u0_, v0_, w0_}, {fc_,gc_,hc_}, {d1_,d2_,d3_}, {hhx_,hhy_}]:= 
	Module[{u1,v1,w1},
		u1= ADIXu@ (dirichletYhalfstep[u0, d1*hhy ] + fc[u0,v0,w0]);
		v1= ADIXv@ (  neumannYhalfstep[v0, d2*hhy ] + gc[u0,v0,w0]);
		w1= ADIXw@ (dirichletYhalfstep[w0, d3*hhy ] + hc[u0,v0,w0]);
		{setedges[ ADIYu@ (dirichletXhalfstep[u1, d1*hhx ] + fc[u1,v1,w1]), First[bdata], a010],
		           ADIYv@ (  neumannXhalfstep[v1, d2*hhx ] + gc[u1,v1,w1]),
		 setedges[ ADIYw@ (dirichletXhalfstep[w1, d3*hhx ] + hc[u1,v1,w1]), Last[bdata],  a010]}
		 ];

RDstepDDD[{u0_, v0_, w0_}, {fc_,gc_,hc_}, {d1_,d2_,d3_}, {hhx_,hhy_}]:= 
	Module[{u1,v1,w1},
		u1= ADIXu@ (dirichletYhalfstep[u0, d1*hhy ] + fc[u0,v0,w0]);
		v1= ADIXv@ (dirichletYhalfstep[v0, d2*hhy ] + gc[u0,v0,w0]);
		w1= ADIXw@ (dirichletYhalfstep[w0, d3*hhy ] + hc[u0,v0,w0]);
		{setedges[ ADIYu@ (dirichletXhalfstep[u1, d1*hhx ] + fc[u1,v1,w1]), First[bdata], a010],
		 setedges[ ADIYv@ (dirichletXhalfstep[v1, d2*hhx ] + gc[u1,v1,w1]), bdata[[2]],   a010],
		 setedges[ ADIYw@ (dirichletXhalfstep[w1, d3*hhx ] + hc[u1,v1,w1]), Last[bdata],  a010] }
		 ];

RDstepDDN[{u0_, v0_, w0_}, {fc_,gc_,hc_}, {d1_,d2_,d3_}, {hhx_,hhy_}]:= 
	Module[{u1,v1,w1},
		u1= ADIXu@ (dirichletYhalfstep[u0, d1*hhy ] + fc[u0,v0,w0]);
		v1= ADIXv@ (dirichletYhalfstep[v0, d2*hhy ] + gc[u0,v0,w0]);
		w1= ADIXw@ (  neumannYhalfstep[w0, d3*hhy ] + hc[u0,v0,w0]);
		{setedges[ ADIYu@ (dirichletXhalfstep[u1, d1*hhx ] + fc[u1,v1,w1]), First[bdata], a010],
		 setedges[ ADIYv@ (dirichletXhalfstep[v1, d2*hhx ] + gc[u1,v1,w1]), bdata[[2]],   a010],
		           ADIYw@ (  neumannXhalfstep[w1, d3*hhx ] + hc[u1,v1,w1])}
		 ];

RDstepDNN[{u0_, v0_, w0_}, {fc_,gc_,hc_}, {d1_,d2_,d3_}, {hhx_,hhy_}]:= 
	Module[{u1,v1,w1},
		u1= ADIXu@ (dirichletYhalfstep[u0, d1*hhy ] + fc[u0,v0,w0]);
		v1= ADIXv@ (  neumannYhalfstep[v0, d2*hhy ] + gc[u0,v0,w0]);
		w1= ADIXw@ (  neumannYhalfstep[w0, d3*hhy ] + hc[u0,v0,w0]);
		{setedges[	ADIYu@ (dirichletXhalfstep[u1, d1*hhx ] + fc[u1,v1,w1]), First[bdata], a010],
		 			ADIYv@ (  neumannXhalfstep[v1, d2*hhx ] + gc[u1,v1,w1]),
					ADIYw@ (  neumannXhalfstep[w1, d3*hhx ] + hc[u1,v1,w1])}
		 ];
		 
RDstepNDN[{u0_, v0_, w0_}, {fc_,gc_,hc_}, {d1_,d2_,d3_}, {hhx_,hhy_}]:= 
	Module[{u1,v1,w1},
		u1= ADIXu@ (  neumannYhalfstep[u0, d1*hhy ] + fc[u0,v0,w0]);
		v1= ADIXv@ (dirichletYhalfstep[v0, d2*hhy ] + gc[u0,v0,w0]);
		w1= ADIXw@ (  neumannYhalfstep[w0, d3*hhy ] + hc[u0,v0,w0]);
		{			ADIYu@ (  neumannXhalfstep[u1, d1*hhx ] + fc[u1,v1,w1]),
		 setedges[	ADIYv@ (dirichletXhalfstep[v1, d2*hhx ] + gc[u1,v1,w1]), bdata[[2]], a010],
					ADIYw@ (  neumannXhalfstep[w1, d3*hhx ] + hc[u1,v1,w1])}
		 ];
		 
RDstepPeriodic3[{u0_, v0_, w0_}, {fc_,gc_,hc_}, {d1_,d2_,d3_}, {hhx_,hhy_}]:= 
	Module[{u1,v1,w1},
		u1= ADIXu@(periodicYhalfstep[u0, d1*hhy ] + fc[u0,v0] );
		v1= ADIXv@(periodicYhalfstep[v0, d2*hhy ] + gc[u0,v0] );
		w1= ADIXw@(periodicYhalfstep[w0, d3*hhy ] + hc[u0,v0] );
	 	{  ADIYu@ (periodicXhalfstep[u1, d1*hhx ] + fc[u1,v1] ),
		   ADIYv@ (periodicXhalfstep[v1, d2*hhx ] + gc[u1,v1] ),
		   ADIYw@ (periodicXhalfstep[w1, d3*hhx ] + hc[u1,v1] ) }
		 ];
		 		
BlockDiag[(mat_)?MatrixQ.., k_] := 
  SparseArray[Flatten[MapThread[Array[List, #1, #2 + 1] & , ({#1, Most[FoldList[Plus, {0, 0}, #1]]} & )[
       PadRight[{{k,k}}, m*n/k, {{k,k}} ] ]], 2] -> PadRight[Flatten[mat], m*n*k, Flatten[mat]]];

sparray[bc_,d_,h_,k_]:=
    Which[
        bc==="Periodic", BlockDiag[
        	Compile[{{kk,_Integer}},
        		Table[ Which[i==j, 1+d*h, Abs[i-j]==1, -.5d*h, Abs[i-j]==k-1, -.5d*h, True, 0],
        	  	{i,kk},{j,kk}] ][k], k
        	  	],
            
        bc==="Dirichlet", BlockDiag[
        	Compile[{{kk,_Integer}},
        		Table[ Which[i==j==1, 1, i==j==k, 1, i==j, 1+d*h, Abs[i-j]==1, -.5d*h, True, 0],
        	  	{i,kk},{j,kk}] ][k], k
        	  	],
				
        bc==="Neumann", BlockDiag[
        	Compile[{{kk,_Integer}},
        		Table[ Which[i==j==1, 1+.5d*h, i==j==k, 1+.5d*h, i==j, 1+d*h, Abs[i-j]==1, -.5d*h, True, 0],
        	  	{i,kk},{j,kk}] ][k], k
        	  	]
        ];

makeLSolvers[{bcu_,bcv_},{du_,dv_},h_]:=
        LinearSolve/@
            { sparray[bcu, du, First@h, n],
			  sparray[bcu, du, 	Last@h, m],
			  sparray[bcv, dv, First@h, n],
			  sparray[bcv, dv, 	Last@h, m] };
            
makeLSolvers[{bcu_,bcv_,bcw_},{du_,dv_,dw_},h_]:=
        LinearSolve/@
            { sparray[bcu, du, First@h, n],
			  sparray[bcu, du,  Last@h, m],
			  sparray[bcv, dv, First@h, n],
			  sparray[bcv, dv,  Last@h, m],
			  sparray[bcw, dw, First@h, n],
			  sparray[bcw, dw,  Last@h, m] };

abortOnError:=(If[System`Dump`TypesetMessagesQ[]&&!MatchQ[#1,_],
     			 System`Dump`TypesetCompact[#1],System`Dump`TextualCompact[#1]];
     			 Print["Use a smaller time step."];
     			 FrontEndExecute[FrontEndToken["EvaluatorAbort"]])&
		
(******* Definition of RDDensityPlots for 2-component systems *********************************)

RDDensityPlots[{f_,g_}, {u_Symbol,v_Symbol}, {u0_?MatrixQ,v0_?MatrixQ}/;Dimensions[u0]===Dimensions[v0],
	{d1_?NonNegative, d2_?NonNegative}, dt_?Positive, nsteps_Integer/;Positive[nsteps], opts___Rule]:=
	
	Module[{uvals, vvals, a,b,c,d,ar, hh, returnlast, wait, bc, step, bdf, a101, ldpopts, grsp, lbls},
		
		ldpopts = Sequence[DisplayFunction->Identity,
					FilterOptions[ListDensityPlot, opts, DeleteCases[Options[RDDensityPlots], AspectRatio->_ ] ] ];
		grsp = GraphicsSpacing/.{opts}/.Options[RDDensityPlots];
		lbls = PlotLabel/.{opts}/.Options[RDDensityPlots];
		If[ Not[MatchQ[lbls,{_,_}]], lbls={None,None}];
		
		returnlast = ReturnLast/.{opts}/.Options[RDDensityPlots];
		bc = ToString/@ Flatten[{BoundaryConditions/.{opts}/.Options[RDDensityPlots]}];
		wait = PlotEvery/.{opts}/.Options[RDDensityPlots];
		bdf = BoundaryFunctions/.{opts}/.Options[RDDensityPlots];

		minu = Min[u0]; maxu = Max[u0];
		minv = Min[v0]; maxv = Max[v0];

		pr = PlotRange/.{opts}/.Options[RDDensityPlots];
		pr = If[ pr===Automatic||NumericQ[First[pr]], {pr, pr}, pr  ]; 
		If[ First[pr]===Automatic, First[pr]= {If[minu<0, 1.25minu, 0], If[maxu<0,0,1.25maxu] } ];
		If[ Last[pr]=== Automatic,  Last[pr]= {If[minv<0, 1.25minv, 0], If[maxv<0,0,1.25maxv] } ];
		
		{m,n} = Dimensions[u0];
		{{a,b},{c,d}} = Domain/.{opts}/.Options[RDDensityPlots]/.{{x1_,Automatic},{y1_,y2_}}->{{x1,(n-1)/(m-1)},{y1,y2}};
		
		If[Not[MatchQ[{{a,b},{c,d}},{{_?NumericQ ..} ..}]], {{a,b},{c,d}} = {{0,(n-1)/(m-1)},{0,1}}];
		
		ar = (d-c)/(b-a);
		
		hh = N@ dt/{((b-a)/(n-1))^2,((d-c)/(m-1))^2};
	
		fc = Compile[{{u,_Real,2},{v,_Real,2}}, .5*dt*f];
		gc = Compile[{{u,_Real,2},{v,_Real,2}}, .5*dt*g];
		
		a010 = Table[ If[ i==1||j==1||i==m||j==n, 0, 1 ], {i,m}, {j,n}];
		a101 = Mod[1+a010,2];
		
		bc= If[Length[bc]===1, Flatten[{bc,bc}], bc];
		
		step= Which[
				bc==={"Periodic","Periodic"},   RDstepPeriodic,
				bc==={"Dirichlet","Dirichlet"}, RDstepDD,
				bc==={"Neumann","Neumann"},     RDstepNN,
				bc==={"Dirichlet","Neumann"},   RDstepDN,
				bc==={"Neumann","Dirichlet"},   RDstepND,
				True, Return[$Failed] ];
						
		If[MemberQ[bc,"Dirichlet"],
				bdata= 
					If[bc[[1]]==="Dirichlet", {a101*Table[First[bdf][x,y], {y,c,d,(d-c)/(m-1)}, {x,a,b,(b-a)/(n-1)} ]}, {{}} ];
				AppendTo[bdata,
					If[bc[[2]]==="Dirichlet",  a101*Table[ Last[bdf][x,y], {y,c,d,(d-c)/(m-1)}, {x,a,b,(b-a)/(n-1)} ], {}] ]
			];
		
		{uLsolvex,uLsolvey,vLsolvex,vLsolvey} = makeLSolvers[bc, {d1,d2}, hh];
	
		{uvals,vvals}=
			N[Which[
			bc==={"Dirichlet","Dirichlet"}, {setedges[u0,First[bdata],a010], setedges[v0,Last[bdata],a010]},
			bc==={"Neumann","Neumann"},     {u0,v0},
			bc==={"Periodic","Periodic"},   {periodicize[u0], periodicize[v0]},
			bc==={"Dirichlet","Neumann"},   {setedges[u0, First[bdata], a010], v0},
			bc==={"Neumann","Dirichlet"},   {u0, setedges[v0, Last[bdata], a010]}
			] ];
	
		Block[{$MessagePrePrint = abortOnError},
		Do[
		{uvals,vvals}= step[{uvals,vvals}, {fc,gc}, {d1,d2}, hh];
		If[Mod[k,Max[wait,2]]==1,
			If[ar <= .6,
			Show[GraphicsArray[{{
			 ListDensityPlot[uvals, PlotRange->First[pr], PlotLabel->First[lbls], ldpopts, AspectRatio->ar ]},
			{ListDensityPlot[vvals, PlotRange->Last[pr],  PlotLabel->Last[lbls],  ldpopts, AspectRatio->ar ]
			}}, GraphicsSpacing->grsp], DisplayFunction->$DisplayFunction, 
						ImageSize:>(ImageSize/.{opts}/.Options[RDDensityPlots]) ],
			Show[GraphicsArray[{
				 ListDensityPlot[uvals, PlotRange->First[pr], PlotLabel->First[lbls], ldpopts, AspectRatio->ar ],
				 ListDensityPlot[vvals, PlotRange->Last[pr],  PlotLabel->Last[lbls],  ldpopts, AspectRatio->ar ]
				}, GraphicsSpacing->grsp], DisplayFunction->$DisplayFunction, 
						ImageSize:>(ImageSize/.{opts}/.Options[RDDensityPlots]) ]
				]],
		{k,1,(nsteps-1)*wait+1}]
		];
		
		If[returnlast, {uvals,vvals}, Null]
		
		] ;

		
(************** Definition of RDSurfacePlots for 2-component systems ***********************)

RDSurfacePlots[{f_,g_}, {u_Symbol,v_Symbol}, {u0_?MatrixQ,v0_?MatrixQ}/;Dimensions[u0]===Dimensions[v0],
	{d1_?NonNegative, d2_?NonNegative}, dt_?Positive, nsteps_Integer/;Positive[nsteps], opts___Rule]:=
	 
	Module[{uvals, vvals, minu, maxu, minv, maxv,  pr, hh, returnlast, wait, bc, bdf, step, a101, lp3opts, grsp, lbls},
	
		lp3opts = Sequence[DisplayFunction->Identity, 
				FilterOptions[ListPlot3D, opts, DeleteCases[Options[RDSurfacePlots], (BoxRatios->_ )|(Ticks->_)]] ];
		grsp = GraphicsSpacing/.{opts}/.Options[RDSurfacePlots];
		lbls = PlotLabel/.{opts}/.Options[RDSurfacePlots];
		If[ Not[MatchQ[lbls,{_,_}]], lbls={None,None}];
		
		returnlast = ReturnLast/.{opts}/.Options[RDSurfacePlots];
		bc = ToString/@Flatten[{BoundaryConditions/.{opts}/.Options[RDSurfacePlots]}];
		wait = PlotEvery/.{opts}/.Options[RDSurfacePlots];
		bdf = BoundaryFunctions/.{opts}/.Options[RDSurfacePlots];

		minu = Min[u0]; maxu=Max[u0];
		minv = Min[v0]; maxv=Max[v0];

		pr = PlotRange/.{opts}/.Options[RDSurfacePlots];
		pr = If[pr===Automatic||NumericQ[First[pr]], {pr, pr}, pr]; 
		If[First[pr]===Automatic, First[pr]= {If[minu<0, 1.25minu, 0], If[maxu<0,0,1.25maxu]}];
		If[Last[pr]=== Automatic,  Last[pr]= {If[minv<0, 1.25minv, 0], If[maxv<0,0,1.25maxv]}];
		
		{m,n} = Dimensions[u0];
		{{a,b},{c,d}} = Domain/.{opts}/.Options[RDDensityPlots]/.{{x1_,Automatic},{y1_,y2_}}->{{x1,(n-1)/(m-1)},{y1,y2}};
		
		If[Not[MatchQ[{{a,b},{c,d}},{{_?NumericQ ..} ..}]], {{a,b},{c,d}} = {{0,(n-1)/(m-1)},{0,1}}];
		
		ar = (d-c)/(b-a);
		
		hh = N@ dt/{((b-a)/(n-1))^2,((d-c)/(m-1))^2};
	
		fc = Compile[{{u,_Real,2},{v,_Real,2}}, .5*dt*f ];
		gc = Compile[{{u,_Real,2},{v,_Real,2}}, .5*dt*g ]; 
		
		a010 = Table[ If[ i==1||j==1||i==m||j==n, 0, 1], {i,m}, {j,n}];
		a101 = Mod[1+a010,2];
		
		bc= If[Length[bc]===1, Append[bc,First[bc]], bc];
		
		step= Which[
				bc==={"Periodic","Periodic"},   RDstepPeriodic,
				bc==={"Dirichlet","Dirichlet"}, RDstepDD,
				bc==={"Neumann","Neumann"},     RDstepNN,
				bc==={"Dirichlet","Neumann"},   RDstepDN,
				bc==={"Neumann","Dirichlet"},   RDstepND,
				True, Return[$Failed] ];
						
		If[MemberQ[bc,"Dirichlet"],
				bdata= 
					If[bc[[1]]==="Dirichlet", {a101*Table[First[bdf][x,y], {y,c,d,(d-c)/(m-1)}, {x,a,b,(b-a)/(n-1)} ]}, {{}} ];
				AppendTo[bdata,
					If[bc[[2]]==="Dirichlet",  a101*Table[ Last[bdf][x,y], {y,c,d,(d-c)/(m-1)}, {x,a,b,(b-a)/(n-1)} ], {}] ]
			];

		{uLsolvex,uLsolvey,vLsolvex,vLsolvey} = makeLSolvers[bc, {d1,d2}, hh];
							
		{uvals,vvals}=
			N[Which[
			bc==={"Dirichlet","Dirichlet"}, {setedges[u0, First[bdata], a010], setedges[v0, Last[bdata], a010] },
			bc==={"Neumann","Neumann"},     {u0, v0},
			bc==={"Periodic","Periodic"},   {periodicize[u0], periodicize[v0]},
			bc==={"Dirichlet","Neumann"},   {setedges[u0, First[bdata], a010], v0},
			bc==={"Neumann","Dirichlet"},   {u0, setedges[v0, Last[bdata], a010]}
			]];
		
		Block[{$MessagePrePrint= abortOnError},
       	Do[
		{uvals,vvals}=step[{uvals,vvals}, {fc,gc}, {d1,d2}, hh];
		If[Mod[k,wait]==1,
			Show[GraphicsArray[{
				ListPlot3D[uvals, PlotRange-> First[pr], PlotLabel->First[lbls], lp3opts, Ticks->{None, None, First[pr]}, BoxRatios->{1, ar, Min[ar,1]} ],
				ListPlot3D[vvals, PlotRange-> Last[pr],  PlotLabel->Last[lbls],  lp3opts, Ticks->{None, None, Last[pr] }, BoxRatios->{1, ar, Min[ar,1]} ]
				}, GraphicsSpacing->grsp], DisplayFunction->$DisplayFunction, Axes->None,
						FilterOptions[Graphics,DeleteCases[Join[{opts},Options[RDSurfacePlots] ],PlotRange->_]]]],
		{k,1,(nsteps-1)*wait+1}]
		];
		
		If[returnlast,{uvals,vvals},Null]
	];
		

(************* Definition of RDDensityPlots for 3-component systems *******************************)
		
RDDensityPlots[{f_,g_,h_}, {u_Symbol,v_Symbol,w_Symbol},
		{u0_?MatrixQ,v0_?MatrixQ,w0_?MatrixQ}/;Dimensions[u0]===Dimensions[v0]===Dimensions[w0],
		{d1_?NonNegative,d2_?NonNegative,d3_?NonNegative}, dt_?Positive, nsteps_Integer/;Positive[nsteps], opts___Rule]:=
		
	Module[{uvals, vvals, wvals, hh, returnlast, wait, bc, step, bdf, a101, pr, ldpopts, grsp, lbls},
					
		ldpopts=Sequence[ DisplayFunction->Identity, 
			FilterOptions[ListDensityPlot, opts, DeleteCases[Options[RDDensityPlots], AspectRatio->_ ] ] ];
		grsp=GraphicsSpacing/.{opts}/.Options[RDDensityPlots];
		lbls = PlotLabel/.{opts}/.Options[RDDensityPlots];
		If[ Not[MatchQ[lbls,{_,_,_}]], lbls={None,None,None}];
		
		returnlast=ReturnLast/.{opts}/.Options[RDDensityPlots];
		bc=ToString/@Flatten[{BoundaryConditions/.{opts}/.Options[RDDensityPlots]}];
		wait=PlotEvery/.{opts}/.Options[RDDensityPlots];
		bdf= BoundaryFunctions/.{opts}/.Options[RDDensityPlots]/.{{f1_,f2_}->{f1,f1,(0&)}};
				
		minu=Min[u0]; maxu=Max[u0];
		minv=Min[v0]; maxv=Max[v0];
		minw=Min[w0]; maxw=Max[w0];

		pr= PlotRange/.{opts}/.Options[RDDensityPlots];
		pr= If[pr===Automatic||NumericQ[pr[[1]]], {pr,pr,pr}, pr]; 
		If[pr[[1]]===Automatic, pr[[1]]={If[minu<0, 1.25minu, 0], If[maxu<0, 0, 1.25maxu]}];
		If[pr[[2]]===Automatic, pr[[2]]={If[minv<0, 1.25minv, 0], If[maxv<0, 0, 1.25maxv]}];
		If[pr[[3]]===Automatic, pr[[3]]={If[minw<0, 1.25minw, 0], If[maxw<0, 0, 1.25maxw]}];	

		{m,n} = Dimensions[u0];
		{{a,b},{c,d}} = Domain/.{opts}/.Options[RDDensityPlots]/.{{x1_,Automatic},{y1_,y2_}}->{{x1,(n-1)/(m-1)},{y1,y2}};
		
		If[Not[MatchQ[{{a,b},{c,d}},{{_?NumericQ ..} ..}]], {{a,b},{c,d}} = {{0,(n-1)/(m-1)},{0,1}}];
		
		ar = (d-c)/(b-a);
		
		hh = N@ dt/{((b-a)/(n-1))^2,((d-c)/(m-1))^2};

		fc=Compile[{{u,_Real,2},{v,_Real,2},{w,_Real,2}}, .5*dt*f];
		gc=Compile[{{u,_Real,2},{v,_Real,2},{w,_Real,2}}, .5*dt*g];
		hc=Compile[{{u,_Real,2},{v,_Real,2},{w,_Real,2}}, .5*dt*h];
		
		a010 = Table[ If[ i==1||j==1||i==m||j==n, 0, 1], {i,m}, {j,n}];
		a101 = Mod[1+a010,2];
		
		bc= If[Length[bc]===1, Flatten@{bc,bc,bc}, bc];
		step= Which[
			bc==={"Dirichlet","Dirichlet","Dirichlet"}, RDstepDDD,
			bc==={"Neumann","Dirichlet","Dirichlet"},   RDstepNDD,
			bc==={"Dirichlet","Neumann","Dirichlet"},   RDstepDND,
			bc==={"Neumann","Neumann","Dirichlet"},     RDstepNND,
			bc==={"Dirichlet","Dirichlet","Neumann"},   RDstepDDN,
			bc==={"Neumann","Dirichlet","Neumann"},     RDstepNDN,
			bc==={"Dirichlet","Neumann","Neumann"},     RDstepDNN,
			bc==={"Neumann","Neumann","Neumann"},       RDstepNNN,
			bc==={"Periodic","Periodic","Periodic"},    RDstepPeriodic3,
			True, Return[$Failed]
			];	
		
		If[MemberQ[bc,"Dirichlet"],
				bdata=If[bc[[1]]==="Dirichlet",{a101*Table[First[bdf][x,y],{y,c,d,(d-c)/(m-1)}, {x,a,b,(b-a)/(n-1)} ]}, {{}} ];
				AppendTo[bdata,
					  If[bc[[2]]==="Dirichlet", a101*Table[bdf[[2]][x,y],  {y,c,d,(d-c)/(m-1)}, {x,a,b,(b-a)/(n-1)} ], {}] ];
				AppendTo[bdata,
					  If[bc[[3]]==="Dirichlet", a101*Table[Last[bdf][x,y], {y,c,d,(d-c)/(m-1)}, {x,a,b,(b-a)/(n-1)} ], {}] ]
			];
		
		{uLsolvex,uLsolvey,vLsolvex,vLsolvey,wLsolvex,wLsolvey} = makeLSolvers[bc, {d1,d2,d3}, hh];	
						
		{uvals,vvals,wvals}=
			N[Which[
			bc==={"Dirichlet","Dirichlet","Dirichlet"},
				{setedges[u0,bdata[[1]],a010], setedges[v0,bdata[[2]],a010], setedges[w0,bdata[[3]],a010]},
			bc==={"Neumann","Neumann","Dirichlet"},
				{u0, v0, setedges[w0,bdata[[3]],a010]},
			bc==={"Dirichlet","Neumann","Dirichlet"},
				{setedges[u0,bdata[[1]],a010], v0, setedges[w0,bdata[[3]],a010]},
			bc==={"Neumann","Dirichlet","Dirichlet"},
				{u0, setedges[v0,bdata[[2]],a010], setedges[w0,bdata[[3]],a010]},
			bc==={"Dirichlet","Dirichlet","Neumann"},
				{setedges[u0,bdata[[1]],a010], setedges[v0,bdata[[2]],a010], w0},
			bc==={"Neumann","Neumann","Neumann"},
				{u0, v0, w0},
			bc==={"Dirichlet","Neumann","Neumann"},
				{setedges[u0,bdata[[1]],a010], v0, w0},
			bc==={"Neumann","Dirichlet","Neumann"},
				{u0, setedges[v0,bdata[[2]], a010], w0},
			bc==={"Periodic","Periodic","Periodic"},
				{periodicize[u0], periodicize[v0], periodicize[w0]}
			] ];

		Block[{$MessagePrePrint= abortOnError},
   		Do[
		{uvals,vvals,wvals}= step[{uvals,vvals,wvals}, {fc,gc,hc}, {d1,d2,d3} , hh];
		If[Mod[k,wait]==1,
			If[ar <= .4,
			Show[GraphicsArray[{
				{ListDensityPlot[uvals, PlotRange->First[pr], PlotLabel->First[lbls], ldpopts, AspectRatio->ar]},
				{ListDensityPlot[vvals, PlotRange->pr[[2]], PlotLabel->lbls[[2]],  ldpopts, AspectRatio->ar]},
				{ListDensityPlot[wvals, PlotRange->Last[pr], PlotLabel->Last[lbls],  ldpopts, AspectRatio->ar]} }, 
						GraphicsSpacing->grsp], DisplayFunction->$DisplayFunction,
						FilterOptions[Graphics,DeleteCases[Join[{opts},Options[RDDensityPlots] ],PlotRange->_]]],
			Show[GraphicsArray[{
			ListDensityPlot[uvals, PlotRange->First[pr], PlotLabel->First[lbls], ldpopts, AspectRatio->ar],
			ListDensityPlot[vvals, PlotRange->pr[[2]], PlotLabel->lbls[[2]],  ldpopts, AspectRatio->ar],
			ListDensityPlot[wvals, PlotRange->Last[pr], PlotLabel->Last[lbls], ldpopts, AspectRatio->ar]
			}, GraphicsSpacing->grsp], DisplayFunction->$DisplayFunction, 
						FilterOptions[Graphics,DeleteCases[Join[{opts},Options[RDDensityPlots] ],PlotRange->_]]]	
				] ],
		{k,1,(nsteps-1)*wait+1}]
		];
		If[returnlast,{uvals,vvals,wvals},Null]];
	

(*************** Definition of RDSurfacePlots for 3-component systems ************************************)
		
RDSurfacePlots[{f_,g_,h_}, {u_Symbol,v_Symbol,w_Symbol},
		{u0_?MatrixQ,v0_?MatrixQ,w0_?MatrixQ}/;Dimensions[u0]===Dimensions[v0]===Dimensions[w0],
		{d1_?NonNegative,d2_?NonNegative,d3_?NonNegative}, dt_?Positive, nsteps_Integer/;Positive[nsteps], opts___Rule]:=
		
	Module[{uvals, vvals, wvals, minu, maxu, minv, maxv, minw, maxw, hh,
				returnlast, wait, bc, step, bdf, a101, pr, lp3opts, grsp, lbls},
		
		lp3opts=Sequence[DisplayFunction->Identity, FilterOptions[ListPlot3D, opts, DeleteCases[Options[RDSurfacePlots], (BoxRatios->_ )|(Ticks->_)]] ];
		grsp= GraphicsSpacing/.{opts}/.{GraphicsSpacing->.1};
		lbls = PlotLabel/.{opts}/.Options[RDSurfacePlots];
		If[ Not[MatchQ[lbls,{_,_,_}]], lbls={None,None,None}];
		
		returnlast = ReturnLast/.{opts}/.Options[RDSurfacePlots];
		bc = ToString/@Flatten[{BoundaryConditions/.{opts}/.Options[RDSurfacePlots]}];
		wait = PlotEvery/.{opts}/.Options[RDSurfacePlots];
		bdf = BoundaryFunctions/.{opts}/.Options[RDSurfacePlots]/.{{f1_,f2_}->{f1,f1,(0&)}};

		minu=Min[u0]; maxu=Max[u0];
		minv=Min[v0]; maxv=Max[v0];
		minw=Min[w0]; maxw=Max[w0];

		pr= PlotRange/.{opts}/.Options[RDSurfacePlots];
		pr= If[pr===Automatic||NumericQ[pr[[1]]], {pr,pr,pr}, pr]; 
		If[pr[[1]]===Automatic, pr[[1]]={If[minu<0, 1.25minu, 0], If[maxu<0, 0, 1.25maxu]}];
		If[pr[[2]]===Automatic, pr[[2]]={If[minv<0, 1.25minv, 0], If[maxv<0, 0, 1.25maxv]}];
		If[pr[[3]]===Automatic, pr[[3]]={If[minw<0, 1.25minw, 0], If[maxw<0, 0, 1.25maxw]}];	
		
		{m,n} = Dimensions[u0];
		{{a,b},{c,d}} = Domain/.{opts}/.Options[RDDensityPlots]/.{{x1_,Automatic},{y1_,y2_}}->{{x1,(n-1)/(m-1)},{y1,y2}};
		
		If[Not[MatchQ[{{a,b},{c,d}},{{_?NumericQ ..} ..}]], {{a,b},{c,d}} = {{0,(n-1)/(m-1)},{0,1}}];
		
		ar = (d-c)/(b-a);
		
		hh = N@ dt/{((b-a)/(n-1))^2,((d-c)/(m-1))^2};
		
		fc=Compile[ {{u,_Real,2},{v,_Real,2},{w,_Real,2}}, .5*dt*f];
		gc=Compile[ {{u,_Real,2},{v,_Real,2},{w,_Real,2}}, .5*dt*g];
		hc=Compile[ {{u,_Real,2},{v,_Real,2},{w,_Real,2}}, .5*dt*h];
		a010=Table[ If[i==1||j==1||i==m||j==n, 0, 1], {i,m}, {j,n}];
		a101=Mod[1+a010,2]; 
				
		bc= If[Length[bc]===1, Flatten@{bc,bc,bc}, bc];
		step= Which[
			bc==={"Dirichlet","Dirichlet","Dirichlet"}, RDstepDDD,
			bc==={"Neumann","Dirichlet","Dirichlet"},   RDstepNDD,
			bc==={"Dirichlet","Neumann","Dirichlet"},   RDstepDND,
			bc==={"Neumann","Neumann","Dirichlet"},     RDstepNND,
			bc==={"Dirichlet","Dirichlet","Neumann"},   RDstepDDN,
			bc==={"Neumann","Dirichlet","Neumann"},     RDstepNDN,
			bc==={"Dirichlet","Neumann","Neumann"},     RDstepDNN,
			bc==={"Neumann","Neumann","Neumann"},       RDstepNNN,
			bc==={"Periodic","Periodic","Periodic"},    RDstepPeriodic3,
			True, Return[$Failed]
			];
		
		If[MemberQ[bc,"Dirichlet"],
				bdata=If[bc[[1]]==="Dirichlet",{a101*Table[First[bdf][x,y],{y,c,d,(d-c)/(m-1)}, {x,a,b,(b-a)/(n-1)} ]}, {{}} ];
				AppendTo[bdata,
					  If[bc[[2]]==="Dirichlet", a101*Table[bdf[[2]][x,y],  {y,c,d,(d-c)/(m-1)}, {x,a,b,(b-a)/(n-1)} ], {}] ];
				AppendTo[bdata,
					  If[bc[[3]]==="Dirichlet", a101*Table[Last[bdf][x,y], {y,c,d,(d-c)/(m-1)}, {x,a,b,(b-a)/(n-1)} ], {}] ]
			];	
						
		{uLsolvex,uLsolvey,vLsolvex,vLsolvey,wLsolvex,wLsolvey} = makeLSolvers[bc, {d1,d2,d3}, hh];	
					
		{uvals,vvals,wvals}=
			N[Which[
			bc==={"Dirichlet","Dirichlet","Dirichlet"},
				{setedges[u0,bdata[[1]],a010], setedges[v0,bdata[[2]],a010], setedges[w0,bdata[[3]],a010]},
			bc==={"Neumann","Neumann","Dirichlet"},
				{u0, v0, setedges[w0,bdata[[3]],a010]},
			bc==={"Dirichlet","Neumann","Dirichlet"},
				{setedges[u0,bdata[[1]],a010], v0, setedges[w0,bdata[[3]],a010]},
			bc==={"Neumann","Dirichlet","Dirichlet"},
				{u0, setedges[v0,bdata[[2]],a010], setedges[w0,bdata[[3]],a010]},
			bc==={"Dirichlet","Dirichlet","Neumann"},
				{setedges[u0,bdata[[1]],a010], setedges[v0,bdata[[2]],a010], w0},
			bc==={"Neumann","Neumann","Neumann"},
				{u0, v0, w0},
			bc==={"Dirichlet","Neumann","Neumann"},
				{setedges[u0,bdata[[1]],a010], v0, w0},
			bc==={"Neumann","Dirichlet","Neumann"},
				{u0, setedges[v0,bdata[[2]], a010], w0},
			bc==={"Periodic","Periodic","Periodic"},
				{periodicize[u0], periodicize[v0], periodicize[w0]}
			] ];
		
		Block[{$MessagePrePrint= abortOnError},
   		Do[
		{uvals,vvals,wvals}=step[{uvals,vvals,wvals},{fc,gc,hc},{d1,d2,d3}, hh];
		If[Mod[k,wait]==1,
			Show[GraphicsArray[{
				ListPlot3D[uvals, PlotRange->pr[[1]], PlotLabel->First[lbls], lp3opts, Ticks->{None, None, pr[[1]]}, BoxRatios->{1, ar, Min[ar,1]}],
				ListPlot3D[vvals, PlotRange->pr[[2]], PlotLabel->lbls[[2]],  lp3opts, Ticks->{None, None, pr[[2]]}, BoxRatios->{1, ar, Min[ar,1]}],
				ListPlot3D[wvals, PlotRange->pr[[3]], PlotLabel->Last[lbls],  lp3opts, Ticks->{None, None, pr[[3]]}, BoxRatios->{1, ar, Min[ar,1]}]
				}, GraphicsSpacing->grsp], DisplayFunction->$DisplayFunction, Axes->None,
				FilterOptions[Graphics,DeleteCases[Join[{opts},Options[RDSurfacePlots] ],PlotRange->_]]]],
		{k,1,(nsteps-1)*wait+1}]
		];
		If[returnlast,{uvals,vvals,wvals},Null]];
		
		
End[]

EndPackage[]

